@@ -504,6 +504,16 @@ def cdefs(features):
 
         '''
 
+    if _at_least(features, 8, 13):
+        code += '''
+            extern "Python" int _marshal_end (VipsTarget*,
+                void*);
+
+            void vips_block_untrusted_set (int state);
+            void vips_operation_block_set (const char *name, int state);
+
+        '''
+
     # we must only define these in API mode ... in ABI mode we need to call
     # these things earlier
     if features['api']:
