@@ -3,10 +3,11 @@
 from __future__ import division
 
 import numbers
+import struct
 
 import pyvips
 from pyvips import ffi, glib_lib, vips_lib, Error, _to_bytes, \
-    _to_string, _to_string_copy, GValue, at_least_libvips
+    _to_string, _to_string_copy, GValue, at_least_libvips, Introspect
 
 
 # either a single number, or a table of numbers
@@ -125,11 +126,131 @@ def _deprecated(note):
     return _dep
 
 
+# Rules for the mapping betweeen numpy typestrings and libvips formats
+# b1: bool. Above u1 so that rev. map is uchar->u1
+TYPESTR_TO_FORMAT = {'|b1': 'uchar',
+                     '|u1': 'uchar',
+                     '|i1': 'char',
+                     '<u2': 'ushort',
+                     '<i2': 'short',
+                     '<u4': 'uint',
+                     '<i4': 'int',
+                     '<f4': 'float',
+                     '<c8': 'complex',
+                     '<f8': 'double',
+                     '<c16': 'dpcomplex'}
+
+# Rules for the mapping between libvips formats and numpy typestrings
+FORMAT_TO_TYPESTR = dict((v, k) for k, v in TYPESTR_TO_FORMAT.items())
+
+# see https://docs.python.org/3/library/struct.html
+FORMAT_TO_PYFORMAT = {'uchar': 'B',
+                      'char': 'c',
+                      'ushort': 'H',
+                      'short': 'h',
+                      'uint': 'I',
+                      'int': 'i',
+                      'float': 'f',
+                      'double': 'd',
+                      'complex': 'f',   # n.b. no support for complex in C
+                      'dpcomplex': 'd'}
+
+
+def _guess_interpretation(bands, format):
+    """Return a best-guess interpretation string based on bands and libvips
+    format.
+
+    Args:
+        bands (int): Number of bands in the image.
+        format (str): libvips format string.
+
+    Returns:
+        str: the best-guess libvips interpretation string.
+
+    This function is a helper for :class:`.Image` creation from arrays when the
+    user specifies 'auto' for `interpretation`.
+
+    The heuristic used here can be defined in the following table::
+
+        | bands | format    | interpretation | reason       |
+        | ----- | --------- | -------------- | ------------ |
+        | 1     | uchar     | b-w            | L            |
+        | 2     | uchar     | b-w            | LA           |
+        | 3     | uchar     | rgb            | RGB          |
+        | 4     | uchar     | rgb            | RGBA         |
+        | 5+    | uchar     | multiband      | spectral     |
+        | 1     | char      | matrix         | conv. kernel |
+        | 2+    | char      | multiband      |              |
+        | 1     | ushort    | grey16         | L            |
+        | 2     | ushort    | grey16         | LA           |
+        | 3     | ushort    | rgb16          | RGB          |
+        | 4     | ushort    | rgb16          | RGBA         |
+        | 5+    | ushort    | multiband      | spectral     |
+        | any   | short     | multiband      |              |
+        | 1     | float     | b-w            | L            |
+        | 2     | float     | b-w            | LA           |
+        | 3     | float     | scrgb          | RGB          |
+        | 4     | float     | scrgb          | RGBA         |
+        | 5+    | float     | multiband      | spectral     |
+        | 1     | double    | b-w            | L            |
+        | 2     | double    | b-w            | LA           |
+        | 3     | double    | scrgb          | RGB          |
+        | 4     | double    | scrgb          | RGBA         |
+        | 5+    | double    | multiband      | spectral     |
+        | any   | complex   | fourier        | FFT          |
+        | any   | dpcomplex | fourier        | FFT          |
+
+    """
+
+    if format not in FORMAT_TO_TYPESTR:
+        raise ValueError('Unknown format: {}'.format(format))
+    if not isinstance(bands, int) or bands < 1:
+        raise ValueError('Number of bands must be a positive integer.')
+
+    # the default
+    interp = 'multiband'
+
+    if format == 'uchar':
+        if bands in [1, 2]:
+            interp = 'b-w'
+        elif bands in [3, 4]:
+            interp = 'rgb'
+    elif format == 'char':
+        if bands == 1:
+            interp = 'matrix'
+    elif format == 'ushort':
+        if bands in [1, 2]:
+            interp = 'grey16'
+        elif bands in [3, 4]:
+            interp = 'rgb16'
+    elif format == 'short':
+        if bands == 1:
+            interp = 'matrix'
+    elif format in ['float', 'double']:
+        if bands in [1, 2]:
+            interp = 'b-w'
+        elif bands in [3, 4]:
+            interp = 'scrgb'
+    elif format in ['complex', 'dpcomplex']:
+        interp = 'fourier'
+
+    return interp
+
+
 # metaclass for Image ... getattr on this implements the class methods
 class ImageType(type):
     def __getattr__(cls, name):
         # logger.debug('ImageType.__getattr__ %s', name)
 
+        # does the method exist in libvips?
+        try:
+            # this will throw an exception if not
+            Introspect.get(name)
+        except Error:
+            # we need to throw this exception for missing methods, eg. numpy
+            # checks for this
+            raise AttributeError
+
         @_add_doc(name)
         def call_function(*args, **kwargs):
             return pyvips.Operation.call(name, *args, **kwargs)
@@ -268,8 +389,8 @@ class Image(pyvips.VipsObject):
                                      string_options=options, **kwargs)
 
     @staticmethod
-    def new_from_array(array, scale=1.0, offset=0.0):
-        """Create an image from a 1D or 2D array.
+    def new_from_list(array, scale=1.0, offset=0.0):
+        """Create an image from a list or list of lists.
 
         A new one-band image with :class:`BandFormat` ``'double'`` pixels is
         created from the array. These image are useful with the libvips
@@ -312,6 +433,146 @@ class Image(pyvips.VipsObject):
 
         return image
 
+    @classmethod
+    def new_from_array(cls, obj, scale=1.0, offset=0.0, interpretation=None):
+        """Create a new Image from a list or an array-like object.
+
+        Array-like objects are those which define `__array_interface__` or
+        `__array__`. For details about the array interface, see `The Array
+        Interface
+        <https://numpy.org/doc/stable/reference/arrays.interface.html>`_.
+
+        If `__array_interface__` is not available, `__array__` is used as a
+        fallback.
+
+        The behavior for input objects with different dimensions is summarized
+        as::
+
+            | array ndim | array shape | Image w | Image h | Image bands |
+            |------------|-------------|---------|---------|-------------|
+            | 0          | ()          | 1       | 1       | 1           |
+            | 1          | (W,)        | W       | 1       | 1           |
+            | 2          | (H, W)      | W       | H       | 1           |
+            | 3          | (H, W, C)   | W       | H       | C           |
+
+
+        Args:
+            obj (list or object width `__array_interface__` or `__array__`):
+                The object to convert to an image.
+
+                If the input object is a list, `Image.new_from_list` is used
+                with the given `scale` and `offset`
+
+                If the input object is an array-like object, a new image is
+                created from the object's data and shape.  The memory is shared
+                except in the following cases:
+
+                - The object's memory is not contiguous.  In this case, a copy
+                  is made by attempting to call the object's `tobytes()` method
+                  or its `tostring()` method.
+
+                - The object is an array of bools, in which case it is
+                  converted to a pyvips uchar image with True values becoming
+                  255 and False values becoming 0.
+
+            scale (float): Default to 1.0. **Ignored for non-list inputs**.
+                What to divide each pixel by after convolution.  Useful for
+                integer convolution masks.
+            offset (float): Default to 0.0.  **Ignored for non-list inputs**.
+                What to subtract from each pixel after convolution.  Useful for
+                integer convolution masks.
+            interpretation (str, optional): **Ignored for list inputs** The
+                libvips interpretation of the array. If None, the
+                interpretation defaults to the pyvips one for
+                `Image.new_from_memory`.
+
+                If 'auto', a heuristic is used to determine a best-guess
+                interpretation as defined in the `_guess_interpretation`
+                function.
+
+                Must be one of None, 'auto', 'error', 'multiband', 'b-w',
+                'histogram', 'xyz', 'lab', 'cmyk', 'labq', 'rgb', 'cmc', 'lch',
+                'labs', 'srgb', 'yxy', 'fourier', 'rgb16', 'grey16', 'matrix',
+                'scrgb', or 'hsv'
+
+        Returns:
+            The new image.
+
+        See Also:
+            :func:`_guess_interpretation`
+        """
+        if isinstance(obj, list):
+            return cls.new_from_list(obj, scale, offset)
+
+        if hasattr(obj, '__array_interface__'):
+            a = obj.__array_interface__
+            shape = a['shape']
+            typestr = a['typestr']
+            ndim = len(shape)
+
+            # strides is optional
+            strides = a.get('strides', None)
+
+            if ndim > 3:
+                raise ValueError('array has more than 3 dimensions')
+            if typestr not in TYPESTR_TO_FORMAT:
+                raise ValueError('conversion from {0} not supported'
+                                 .format(typestr))
+
+            if ndim == 0:
+                width = 1
+                height = 1
+                bands = 1
+            elif ndim == 1:
+                width = shape[0]
+                height = 1
+                bands = 1
+            elif ndim == 2:
+                height, width = shape
+                bands = 1
+            elif ndim == 3:
+                height, width, bands = shape
+
+            format = TYPESTR_TO_FORMAT[typestr]
+
+            if strides is None and hasattr(obj, 'data'):
+                data = obj.data
+            else:
+                # To obtain something with a contiguous memory layout
+                if hasattr(obj, 'tobytes'):
+                    data = obj.tobytes()
+                elif hasattr(obj, 'tostring'):
+                    data = obj.tostring()
+                else:
+                    raise TypeError('object has no .tobytes or .tostring')
+
+            im = cls.new_from_memory(
+                data,
+                width,
+                height,
+                bands,
+                format
+            )
+
+            if typestr == '|b1':
+                # special case for bool: true in vips is uchar(255)
+                im = im.ifthenelse(255, 0)
+
+            if interpretation is not None:
+                if interpretation == 'auto':
+                    interpretation = _guess_interpretation(bands, format)
+
+                im = im.copy(interpretation=interpretation)
+
+            return im
+
+        elif hasattr(obj, '__array__'):
+            # make it into something that *does* define __array_interface__
+            return cls.new_from_array(obj.__array__())
+        else:
+            raise TypeError('does not define __array_interface__ '
+                            'or __array__')
+
     @staticmethod
     def new_from_memory(data, width, height, bands, format):
         """Wrap an image around a memory array.
@@ -347,7 +608,6 @@ class Image(pyvips.VipsObject):
             :class:`.Error`
 
         """
-
         format_value = GValue.to_enum(GValue.format_type, format)
         pointer = ffi.from_buffer(data)
         # py3:
@@ -436,7 +696,6 @@ class Image(pyvips.VipsObject):
             :class:`.Error`
 
         """
-
         vi = vips_lib.vips_image_new_temp_file(_to_bytes(format))
         if vi == ffi.NULL:
             raise Error('unable to make temp file')
@@ -461,7 +720,6 @@ class Image(pyvips.VipsObject):
             :class:`.Error`
 
         """
-
         pixel = (Image.black(1, 1) + value).cast(self.format)
         image = pixel.embed(0, 0, self.width, self.height,
                             extend='copy')
@@ -680,7 +938,6 @@ class Image(pyvips.VipsObject):
             :class:`.Error`
 
         """
-
         psize = ffi.new('size_t *')
         pointer = vips_lib.vips_image_write_to_memory(self.pointer, psize)
         if pointer == ffi.NULL:
@@ -709,6 +966,23 @@ class Image(pyvips.VipsObject):
         if result != 0:
             raise Error('unable to write to image')
 
+    def invalidate(self):
+        """Drop caches on an image, and any downstream images.
+
+        This method drops all pixel caches on an image and on all downstream
+        images. Any operations which depend on this image, directly or
+        indirectly, are also dropped from the libvips operation cache.
+
+        This method can be useful if you wrap a libvips image around an area
+        of memory with :meth:`.new_from_memory` and then change some bytes
+        without libvips knowing.
+
+        Returns:
+            None
+
+        """
+        vips_lib.vips_image_invalidate_all(self.pointer)
+
     def set_progress(self, progress):
         """Enable progress reporting on an image.
 
@@ -746,7 +1020,6 @@ class Image(pyvips.VipsObject):
             None
 
         """
-
         # on libvips before 8.5, property types must be fetched separately,
         # since built-in enums were reported as ints
         if not at_least_libvips(8, 5):
@@ -775,7 +1048,6 @@ class Image(pyvips.VipsObject):
             :class:`.Error`
 
         """
-
         # with old libvips, we must fetch properties (as opposed to
         # metadata) via VipsObject
         if not at_least_libvips(8, 5):
@@ -798,7 +1070,6 @@ class Image(pyvips.VipsObject):
             [string]
 
         """
-
         names = []
 
         if at_least_libvips(8, 5):
@@ -832,7 +1103,6 @@ class Image(pyvips.VipsObject):
             None
 
         """
-
         gv = GValue()
         gv.set_type(gtype)
         gv.set(value)
@@ -877,13 +1147,144 @@ class Image(pyvips.VipsObject):
             None
 
         """
-
         return vips_lib.vips_image_remove(self.pointer, _to_bytes(name)) != 0
 
+    def tolist(self):
+        """Return a single-band image as a list of lists.
+
+        Returns:
+            list of lists of values
+
+        """
+        if not self.bands == 1:
+            raise NotImplementedError('tolist only implemented for ' +
+                                      'single-band images')
+
+        is_complex = self.format in ['complex', 'dpcomplex']
+
+        row_els = self.width if not is_complex else 2 * self.width
+
+        rowfmt = '{0}{1}'.format(row_els, FORMAT_TO_PYFORMAT[self.format])
+        buf = self.write_to_memory()
+
+        lst = [list(row) for row in struct.iter_unpack(rowfmt, buf)]
+
+        if is_complex:
+            lst = [[r[i] + 1j*r[i+1] for i in range(0, row_els, 2)]
+                   for r in lst]
+
+        return lst
+
+    # Following will be relevant if we expose the buffer interface:
+    # @property
+    # def __array_interface__(self):
+    #     """Return a numpy-standard __array_interface__ dictionary.
+    #
+    #     This is used by some libraries (PIL, e.g.) to interpret the image
+    #     as an array.
+    #
+    #     See https://numpy.org/doc/stable/reference/arrays.interface.html for
+    #     more info.
+    #
+    #     """
+    #     if self.bands == 1:
+    #         shape = (self.height, self.width)
+    #     else:
+    #         shape = (self.height, self.width, self.bands)
+    #
+    #     typestr = FORMAT_TO_TYPESTR[self.format]
+    #
+    #     interface = {
+    #         'shape': shape,
+    #         'strides': None,
+    #         'descr': [('', typestr)],
+    #         'typestr': typestr,
+    #         'version': 3
+    #     }
+    #
+    #     return interface
+
+    def __array__(self, dtype=None):
+        """Conversion to a NumPy array.
+
+        Args:
+            dtype (str or numpy dtype, optional) The dtype to use for the
+                numpy array. If None, the default dtype of the image is used
+                as defined the global `FORMAT_TO_TYPESTR` dictionary.
+
+        Returns:
+            numpy.ndarray: The array representation of the image.
+                * Single-band images lose their channel axis.
+                * Single-pixel single-band images are converted to a 0D array.
+
+        See https://numpy.org/devdocs/user/basics.dispatch.html for more
+        details.
+
+        This enables a :class:`Image` to be used where a numpy array is
+        expected, including in plotting and conversion to other array
+        container types (pytorch, JAX, dask, etc.), for example.
+
+        `numpy` is a runtime dependency of this function.
+
+        See Also `Image.new_from_array` for the inverse operation. #TODO
+        """
+        import numpy as np
+
+        arr = (
+            np.frombuffer(self.write_to_memory(),
+                          dtype=FORMAT_TO_TYPESTR[self.format])
+            .reshape(self.height, self.width, self.bands)
+        )
+
+        if self.bands == 1:
+            # flatten single-band images
+            arr = arr.squeeze(axis=-1)
+
+            if self.width == 1 and self.height == 1:
+                # flatten single-pixel single-band images
+                arr = arr.squeeze(axis=(0, 1))
+
+        if dtype is not None:
+            arr = arr.astype(dtype)
+
+        return arr
+
+    def numpy(self, dtype=None):
+        """Convenience function to allow numpy conversion to be at the end
+        of a method chain.
+
+        This mimics the behavior of pytorch: ``arr = im.op1().op2().numpy()``
+
+        numpy is a runtime dependency of this function.
+
+        Args:
+            dtype (str or numpy dtype): The dtype to use for the numpy array.
+                If None, the default dtype of the image is used.
+
+        Returns:
+            numpy.ndarray: The image as a numpy array.
+
+        See Also:
+
+            - :meth:`.__array__`
+            - `FORMAT_TO_TYPESTR`: Global dictionary mapping libvips format
+              strings to numpy dtype strings.
+        """
+        return self.__array__(dtype=dtype)
+
     def __repr__(self):
-        return ('<pyvips.Image {0}x{1} {2}, {3} bands, {4}>'.
-                format(self.width, self.height, self.format, self.bands,
-                       self.interpretation))
+        if (
+            self.interpretation == "matrix"
+            and self.width < 20
+            and self.height < 20
+            and self.bands == 1
+        ):
+            array = self.tolist()
+            return repr(array)
+        else:
+            return ('<pyvips.Image {0}x{1} {2}, {3} bands, {4}>'.
+                    format(self.width, self.height, self.format, self.bands,
+                           self.interpretation))
 
     def __getattr__(self, name):
         """Divert unknown names to libvips.
@@ -913,7 +1314,6 @@ class Image(pyvips.VipsObject):
             :class:`.Error`
 
         """
-
         # logger.debug('Image.__getattr__ %s', name)
 
         # scale and offset have default values
@@ -933,6 +1333,15 @@ class Image(pyvips.VipsObject):
         if super(Image, self).get_typeof(name) != 0:
             return super(Image, self).get(name)
 
+        # does the method exist in libvips?
+        try:
+            # this will throw an exception if not
+            Introspect.get(name)
+        except Error:
+            # we need to throw this exception for missing methods, eg. numpy
+            # checks for this
+            raise AttributeError
+
         @_add_doc(name)
         def call_function(*args, **kwargs):
             return pyvips.Operation.call(name, self, *args, **kwargs)
@@ -966,46 +1375,98 @@ class Image(pyvips.VipsObject):
         pass
 
     def __getitem__(self, arg):
-        """Overload []
+        """Overload [] to pull out band elements from an image.
 
-        Use [] to pull out band elements from an image. You can use all the
-        usual slicing syntax, for example::
+        The following arguments types are accepted:
+
+        * int::
 
             green = rgb_image[1]
 
-        Will make a new one-band image from band 1 (the middle band). You can
-        also write::
+          Will make a new one-band image from band 1 (the middle band).
+
+        * slice::
 
             last_two = rgb_image[1:]
             last_band = rgb_image[-1]
             middle_few = multiband[1:-2]
+            reversed  = multiband[::-1]
+            every_other = multiband[::2]
+            other_every_other = multiband[1::2]
+
+        * list of int::
+
+            # list of integers
+            desired_bands = [1, 2, 2, -1]
+            four_band = multiband[desired_bands]
+
+        * list of bool::
+
+            wanted_bands = [True, False, True, True, False]
+            three_band = five_band[wanted_bands]
+
+        In the case of integer or slice arguments, the semantics of slicing
+        exactly match those of slicing `range(self.bands)`.
+
+        In the case of list arguments, the semantics match those of numpy's
+        extended slicing syntax. Thus, lists of booleans must have as many
+        elements as there are bands in the image.
 
         """
-
-        if isinstance(arg, slice):
-            i = 0
-            if arg.start is not None:
-                i = arg.start
-
-            n = self.bands - i
-            if arg.stop is not None:
-                if arg.stop < 0:
-                    n = self.bands + arg.stop - i
-                else:
-                    n = arg.stop - i
-        elif isinstance(arg, int):
-            i = arg
+        if isinstance(arg, int):
+            # raises IndexError for bad integer indices
+            i = range(self.bands)[arg]
             n = 1
+        elif isinstance(arg, slice):
+            band_seq = range(self.bands)[arg]
+            if len(band_seq) == 0:
+                raise IndexError('empty slice')
+
+            i = band_seq[0]
+            n = len(band_seq)
+        elif isinstance(arg, list):
+            if not arg:
+                raise IndexError('empty list')
+
+            # n.b. We don't use isinstance because isinstance(True, int)
+            # is True!
+            if not (all(type(x) == int for x in arg) or
+                    all(type(x) == bool for x in arg)):
+                raise IndexError('list must contain only ints or only bools')
+
+            if isinstance(arg[0], bool):
+                if len(arg) != self.bands:
+                    raise IndexError('boolean index must have same ' +
+                                     'length as bands')
+
+                band_seq = [i for i, x in enumerate(arg) if x]
+                n = len(band_seq)
+                if n == 0:
+                    raise IndexError('empty boolean index')
+                if n == 1:
+                    i = band_seq[0]
+            else:
+                all_bands = range(self.bands)
+                # raises IndexError for bad int indices
+                band_seq = [all_bands[i] for i in arg]
+
+                n = len(band_seq)
+                if n == 1:
+                    i = band_seq[0]
         else:
-            raise TypeError
+            raise IndexError('argument must be an int, slice, ' +
+                             'list of ints, or list of bools')
 
-        if i < 0:
-            i = self.bands + i
-
-        if i < 0 or i >= self.bands:
-            raise IndexError
-
-        return self.extract_band(i, n=n)
+        if n == 1:
+            # int or 1-element selection
+            return self.extract_band(i)
+        elif isinstance(arg, slice) and arg.step == 1:
+            # sequential multi-band slice
+            return self.extract_band(i, n=n)
+        else:
+            # nonsequential slice, list of ints, or list of bools
+            bands = [self.extract_band(x) for x in band_seq]
+            return bands[0].bandjoin(bands[1:])
 
     # overload () to mean fetch pixel
     def __call__(self, x, y):
@@ -1022,7 +1483,6 @@ class Image(pyvips.VipsObject):
             :class:`.Error`
 
         """
-
         return self.getpoint(x, y)
 
     # operator overloads
@@ -1187,6 +1647,10 @@ class Image(pyvips.VipsObject):
         if not isinstance(other, list):
             other = [other]
 
+        # guard against empty list
+        if not other:
+            return self
+
         # if [other] is all numbers, we can use bandjoin_const
         non_number = next((x for x in other
                            if not isinstance(x, numbers.Number)),
@@ -1451,7 +1915,6 @@ class Image(pyvips.VipsObject):
             :class:`.Error`
 
         """
-
         return pyvips.Operation.call('scale', self, **kwargs)
 
 
